import base64
import dataclasses
import io
import json
import os
import random
from dataclasses import dataclass
from typing import Optional

from matplotlib import pyplot as plt
from matplotlib.colors import Colormap
import numpy as np
from IPython.display import DisplayObject

VAPE_URL = os.environ.get("VAPE_URL", "https://keksboter.github.io/vape4d")


_TEMPLATE_IFRAME = """
    <div>
        <iframe id="{canvas_id}" src="{viewer_url}/index.html?inline" width="{canvas_width}" height="{canvas_height}" frameBorder="0" sandbox="allow-same-origin allow-scripts"></iframe>
    </div>
    <script>

        window.addEventListener(
            "message",
            (event) => {{
                if (event.data !== "ready") {{
                    return;
                }}
                let data_decoded = Uint8Array.from(atob("{data_code}"), c => c.charCodeAt(0));
                let cmap_decoded = Uint8Array.from(atob("{cmap_code}"), c => c.charCodeAt(0));
                const iframe = document.getElementById("{canvas_id}");
                if (iframe === null) return;
                iframe.focus();
                iframe.contentWindow.postMessage({{
                    volume: data_decoded,
                    cmap: cmap_decoded,
                    settings: {settings_json}
                }},
                "*");
            }},
            false,
        );
    </script>
"""


@dataclass(unsafe_hash=True)
class ViewerSettings:
    width: int
    height: int
    background_color: tuple
    show_colormap_editor: bool
    show_volume_info: bool
    vmin: Optional[float]
    vmax: Optional[float]
    distance_scale: float
    duration: Optional[float] = None


def viewer(
    data: np.ndarray,
    colormap: Optional[Colormap] = None,
    width: int = 800,
    height: int = 600,
    background_color=(0.0, 0.0, 0.0, 1.0),
    show_colormap_editor=False,
    show_volume_info=False,
    vmin=None,
    vmax=None,
    distance_scale=1.0,
    duration=None,
):
    """_summary_

    Args:
        data (np.ndarray): volume data of shape [T,C, D, H, W]
        colormap (Optional[Colormap], optional): _description_. Defaults to matplotlib default colormap.
        width (int, optional): viewer width. Defaults to 800.
        height (int, optional): viewer height. Defaults to 600.
        background_color (tuple, optional): background color in renderer. Defaults to black.
        show_colormap_editor (bool, optional): show the transfer function editor. Defaults to False.
        show_volume_info (bool, optional): show the volume info window. Defaults to False.
        vmin (float, optional): all values in data are clamped to this value. Defaults to minimum value in data.
        vmax (float, optional):  all values in data are clamped to this value. Defaults to maximum value in data.
        distance_scale (float, optional): distance scale used for rendering. Defaults to 1.0.
        duration (_type_, optional): duration of one animation cycle. Defaults to 5 seconds.
    """
    if colormap is None:
        colormap = plt.get_cmap()
    return VolumeViewer(
        data,
        colormap,
        ViewerSettings(
            width,
            height,
            background_color,
            show_colormap_editor,
            show_volume_info,
            vmin,
            vmax,
            distance_scale,
            duration,
        ),
    )


class VolumeViewer(DisplayObject):
    def __init__(self, data: np.ndarray, colormap: Colormap, settings: ViewerSettings):
        super(VolumeViewer, self).__init__(
            data={"volume": data, "cmap": colormap, "settings": settings}
        )

    def _repr_html_(self):

        data = self.data["volume"]
        colormap = self.data["cmap"]
        settings = self.data["settings"]
        buffer = io.BytesIO()
        np.save(buffer, data.astype(np.float32))
        data_code = base64.b64encode(buffer.getvalue())

        buffer2 = io.BytesIO()
        colormap_data = colormap(np.linspace(0, 1, 256)).astype(np.float32)
        np.save(buffer2, colormap_data)
        cmap_code = base64.b64encode(buffer2.getvalue())

        canvas_id = f"vape4d_canvas_{str(random.randint(0,2**32))}"
        html_code = _TEMPLATE_IFRAME.format(
            canvas_id=canvas_id,
            data_code=data_code.decode("utf-8"),
            cmap_code=cmap_code.decode("utf-8"),
            canvas_width=settings.width,
            canvas_height=settings.height,
            settings_json=json.dumps(dataclasses.asdict(settings)),
            viewer_url=VAPE_URL,
        )
        return html_code

    def __html__(self):
        return self._repr_html_()



# %%
import json
from matplotlib import pyplot as plt
from os import makedirs
from matplotlib.colors import LinearSegmentedColormap, ListedColormap, Colormap
import numpy as np


def export_colormaps(cm_module, folder: str):
    makedirs(folder, exist_ok=True)
    for name, cmap in vars(cm_module).items():
        if not isinstance(cmap, Colormap):
            continue
        if name.endswith("_r"):
            continue
        if name.startswith("cmr."):
            name = name[4:]
        if isinstance(cmap, ListedColormap):
            data = cmap(np.linspace(0, 1, 256)).astype(np.float32)
            np.save("{}/{}.npy".format(folder, name), data)
        elif isinstance(cmap, LinearSegmentedColormap):
            segments = cmap._segmentdata

            if any(
                not isinstance(segments[c], np.ndarray)
                and not isinstance(segments[c], list)
                and not isinstance(segments[c], tuple)
                for c in ["red", "green", "blue"]
            ):
                data = cmap(np.linspace(0, 1, 256)).astype(np.float32)
                np.save("{}/{}.npy".format(folder, name), data)
            else:
                channels = {
                    c: (
                        segments[c].tolist()
                        if isinstance(segments[c], np.ndarray)
                        else segments[c]
                    )
                    for c in ["red", "green", "blue"]
                }
                if "alpha" in channels:
                    channels["alpha"] = segments["alpha"].tolist()
                with open("{}/{}.json".format(folder, name), "w") as f:
                    json.dump(
                        channels,
                        f,
                        indent=2,
                    )
        else:
            raise TypeError(
                "cmap must be either a ListedColormap or a LinearSegmentedColormap"
            )


export_colormaps(plt.cm, "colormaps/matplotlib")

import seaborn as sns

export_colormaps(sns.cm, "colormaps/seaborn")

import cmasher

export_colormaps(cmasher.cm, "colormaps/cmasher")



from .render import render
from . import utils

try:
    # check if ipython is available
    import IPython as _
    from .viewer import viewer, ViewerSettings, VolumeViewer
except ImportError:
    pass

from . import vape4d as vape_py

__doc__ = vape_py.__doc__
if hasattr(vape_py, "__all__"):
    __all__ = vape_py.__all__



from copy import deepcopy
from matplotlib.colors import Colormap, LinearSegmentedColormap, ListedColormap
import numpy as np


def diverging_alpha(cmap: Colormap) -> Colormap:
    """changes the alpha channel of a colormap to be diverging (0->1, 0.5 > 0, 1->1)

    Args:
        cmap (Colormap): colormap

    Returns:
        Colormap: new colormap
    """
    cmap = cmap.copy()
    if isinstance(cmap, ListedColormap):
        cmap.colors = deepcopy(cmap.colors)
        for i, a in enumerate(cmap.colors):
            a.append(2 * abs(i / cmap.N - 0.5))
    elif isinstance(cmap, LinearSegmentedColormap):
        cmap._segmentdata["alpha"] = np.array(
            [[0.0, 1.0, 1.0], [0.5, 0.0, 0.0], [1.0, 1.0, 1.0]]
        )
    else:
        raise TypeError(
            "cmap must be either a ListedColormap or a LinearSegmentedColormap"
        )
    return cmap


def linear_increasing_alpha(cmap: Colormap) -> Colormap:
    """changes the alpha channel of a colormap to be linear (0->0, 1->1)

    Args:
        cmap (Colormap): colormap

    Returns:
        Colormap: new colormap
    """
    cmap = cmap.copy()
    if isinstance(cmap, ListedColormap):
        cmap.colors = cmap.colors.copy()
        for i, a in enumerate(cmap.colors):
            a.append(i / (cmap.N - 1))
    elif isinstance(cmap, LinearSegmentedColormap):
        cmap._segmentdata["alpha"] = np.array([[0.0, 0.0, 0.0], [1.0, 1.0, 1.0]])
    else:
        raise TypeError(
            "cmap must be either a ListedColormap or a LinearSegmentedColormap"
        )
    return cmap



from typing import Optional, Union,Tuple,List
from matplotlib import pyplot as plt
from matplotlib.colors import Colormap
import numpy as np
from . import vape4d


def render(
    volume: np.ndarray,
    cmap: Optional[Colormap] = None,
    time: Optional[Union[float, List[float]]] = 0.0,
    width: int = 1024,
    height: int = 1024,
    background: Tuple[float, float, float, float] = (0, 0, 0, 1),
    vmin: Optional[float] = None,
    vmax: Optional[float] = None,
    distance_scale: float = 1.0,
    spatial_interpolation: str = "linear",
    temporal_interpolation: str = "linear",
    camera_angle: Optional[Tuple[float, float]] = None,
) -> np.ndarray:
    """renders a single or multiple images of a volume

    Args:
        volume (np.ndarray): volume data of shape [N, D, H, W]
        cmap (Colormap): colormap to use for rendering. Defaults to matplotlib's default colormap.
        time (Union[float, list[float]]): if a single value is given, a single image is rendered at that time. If a list of values is given, a video is rendered with the given times.
        width (int, optional): image width. Defaults to 1024.
        height (int, optional): image height. Defaults to 1024.
        background (tuple[float, float, float, float], optional): background color. Defaults to (0, 0, 0, 1).
        vmin (Optional[float], optional): minimum value for colormap. defaults to minimum value in volume.
        vmax (Optional[float], optional): maximum value for colormap. defaults to maximum value in volume.
        distance_scale (float, optional): distance scale for rendering. A larger value makes everything more opaque. Defaults to 1.0.
        spatial_interpolation (str, optional): interpolation in space. Linear or Nearest. Defaults to "linear".
        temporal_interpolation (str, optional): interpolation in time. Linear or Nearest. Defaults to "linear".
        camera_angle (Optional[Tuple[float, float]], optional): camera angle for rendering in spherical coordinates (polar,azimuthal angle ). Defaults to None.

    Returns:
        np.ndarray: [T, H, W, 4] if time is a list, [H, W, 4] if time is a single value
    """

    if cmap is None:
        cmap = plt.get_cmap()

    if volume.ndim == 5:
        # check if we have a single channel
        if volume.shape[1] != 1:
            raise ValueError("only one channel supported")
    elif volume.ndim == 4:
        pass
    elif volume.ndim == 3:
        # add one channel
        volume = volume[:, None]
    else:
        raise ValueError(
            "volume must have shape [T,1, D, H, W], [T, D, H, W] or [D,H,W] "
        )

    colormap_data = cmap(np.linspace(0, 1, 256)).astype(np.float32)

    if isinstance(time, np.ndarray):
        time = time.tolist()

    if not isinstance(time, list):
        time = [time]

    frames = vape4d.render_video(
        np.ascontiguousarray(volume).astype(np.float16),
        colormap_data,
        width,
        height,
        time,
        background,
        distance_scale,
        vmin,
        vmax,
        spatial_interpolation,
        temporal_interpolation,
        camera_angle,
    )
    if len(time) == 1:
        return frames[0]
    return frames
